@@ -15,7 +15,6 @@
 /																	   /
 /																	   /
 /---------------------------------------------------------------------*/
-#define		  Assigned_Revision 950802
 /*-------------------------[ Revision History ]------------------------/
 / Revision 1.0.0  :	 Original Code by Jeffry J. Brickley			   /
 /		   1.1.0  : added header capability, JJB, 950802
@@ -33,11 +32,10 @@
 /					fcopy checks for incomplete write (disk full)	Joy/
 /          1.8.0  : C++ version, Boolean tag replaced with C++ bool,   /
 /                   Int_Tag added                                   Joy/
+/          1.8.1  : trim() - memmove() for overlapping memory. Thanks  /
+/                   to Thothy                                       Joy/
+/          1.9.0  : compress/expand Path                            Joy/
 /------------------------------------------------------------------->>*/
-/*	Please keep revision number current.							  */
-#define		  REVISION_NO "1.8.0"
-
-extern "C" {
 
 #define ERROR	-1
 
@@ -75,12 +73,41 @@ extern "C" {
 /
 /---------------------------------------------------------------------*/
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
 #include "cfgopts.h"
+#include "debug.h"
 
-char	*trim(char *buffer)
+ConfigOptions::ConfigOptions(const char *cfgfile, const char *home, const char *data)
+{
+	config_file = cfgfile;
+
+	// prepare home_folder with trailing slash
+	safe_strncpy(home_folder, home, sizeof(home_folder));
+	int hflen = strlen(home_folder);
+	if (hflen > 0 && home_folder[hflen-1] != '/' && home_folder[hflen-1] != '\\')
+		strcat(home_folder, "/");
+
+	// prepare data_folder with trailing slash
+	safe_strncpy(data_folder, data, sizeof(data_folder));
+	int dflen = strlen(data_folder);
+	if (dflen > 0 && data_folder[dflen-1] != '/' && data_folder[dflen-1] != '\\')
+		strcat(data_folder, "/");
+	
+	// prepare config_folder with trailing slash
+	safe_strncpy(config_folder, config_file, sizeof(config_folder));
+	char *slash = strrchr(config_folder, '/');
+	char *alt_slash = strrchr(config_folder, '\\');
+	if (slash != NULL) {
+		slash[1] = '\0';
+	}
+	else if (alt_slash != NULL) {
+		alt_slash[1] = '\0';
+	}
+	else {
+		config_folder[0] = '\0';
+	}
+}
+
+char * ConfigOptions::trim(char *buffer)
 {
 #define	SPACE	' '
 #define	TABULA	'\t'
@@ -94,7 +121,7 @@ char	*trim(char *buffer)
 
 		if (i > 0 && i < linelen) {
 			linelen -= i;
-			memcpy(buffer, buffer + i, linelen);	/* mezery zleva pryc */
+			memmove(buffer, buffer + i, linelen);	/* trim spaces on left */
 		}
 
 		for (i = linelen; i > 0; i--) {
@@ -103,14 +130,14 @@ char	*trim(char *buffer)
 				break;
 		}
 
-		buffer[i] = '\0';						/* mezery zprava pryc */
+		buffer[i] = '\0';						/* trim spaces on right */
 	}
 
 	return buffer;
 }
 
 
-char	*strip_comment(char *line)
+char * ConfigOptions::strip_comment(char *line)
 {
 #define	REM1	'#'
 #define REM2	';'
@@ -148,7 +175,7 @@ char	*strip_comment(char *line)
 /	  char *dest			  ; Destination file name
 /	  char *source			  ; Source file name
 /-------------------------------------------------------------------<<*/
-long fcopy(const char *dest, const char *source)
+long ConfigOptions::fcopy(const char *dest, const char *source)
 {
 	FILE * d, *s;
 	char	*buffer;
@@ -194,8 +221,84 @@ long fcopy(const char *dest, const char *source)
 	return totcount;
 }
 
+void ConfigOptions::expand_path(char *dest, const char *path, unsigned short buf_size)
+{
+	dest[0] = '\0';
+
+	if ( !strlen(path) )
+		return;
+
+	size_t prefixLen = 0;
+
+	if ( path[0] == '~' && (path[1] == '/' || path[1] == '\\') ) {
+		safe_strncpy(dest, home_folder, buf_size);
+		path+=2;
+	} else if (path[0] == '*' && (path[1] == '/' || path[1] == '\\') ) {
+		safe_strncpy(dest, data_folder, buf_size);
+		path+=2;
+	} else if ( path[0] != '/' && path[0] != '\\' && path[1] != ':' ) {
+		safe_strncpy(dest, config_folder, buf_size);
+	}
+	else {
+		safe_strncpy(dest, path, buf_size);
+		return;
+	}
+
+	prefixLen = strlen( dest );
+
+	if ( buf_size >= prefixLen + strlen(path) ) {
+		memmove( dest + prefixLen, path, strlen(path)+1 );	
+	} else {
+		panicbug("Error - config entry size is insufficient");
+	}
+}
+
+void ConfigOptions::compress_path(char *dest, char *path, unsigned short buf_size)
+{
+	dest[0] = '\0';
+
+	if ( !strlen(path) ) {
+		return;
+	}
+
+	size_t prefixLen = 0;
+	const char *replacement = NULL;
+
+	safe_strncpy(dest, config_folder, buf_size);
+	prefixLen = strlen(dest);
+	if (prefixLen && strncmp(path, dest, prefixLen) == 0) {
+		replacement = "";
+		D(bug("%s matches %.*s", path, prefixLen, dest));
+	} 
+	else {
+		safe_strncpy(dest, data_folder, buf_size);
+		prefixLen = strlen(dest);
+		if (prefixLen && strncmp(path, dest, prefixLen) == 0) {
+			replacement = "*";
+			--prefixLen;
+			D(bug("%s matches %.*s", path, prefixLen, dest));
+		} 
+		else {
+			safe_strncpy(dest, home_folder, buf_size);
+			prefixLen = strlen(dest);
+			if (prefixLen && strncmp(path, dest, prefixLen) == 0) {
+				replacement = "~";
+				--prefixLen;
+				D(bug("%s matches %.*s", path, prefixLen, dest));
+			}
+		}
+	}
+
+	if (replacement) {
+		int len1 = strlen(replacement);
+		strcpy(dest, replacement);
+		safe_strncpy(dest+len1, path+prefixLen, buf_size - len1);
+	}
+	else {
+		safe_strncpy(dest, path, buf_size);
+	}
+}
 
-static char	line[32768];
 
 /*---------------------------------------------------------------------/
 /	reads from an input configuration (INI) file.
@@ -204,18 +307,17 @@ static char	line[32768];
 / return value:
 /	  int					  ; number of records read or -1 on error
 / parameters:
-/	  char *filename		  ; filename of INI style file
 /	  struct Config_Tag configs[]; Configuration structure
 /	  char *header			  ; INI header name (i.e. "[TEST]")
 /-------------------------------------------------------------------<<*/
-int	input_config(const char *filename, struct Config_Tag configs[], char *header)
+int	ConfigOptions::input_config(struct Config_Tag configs[], const char *header)
 {
 	struct Config_Tag *ptr;
 	int	count = 0, lineno = 0, temp;
 	FILE * file;
 	char	*fptr, *tok, *next;
 
-	file = fopen(filename, "rt");
+	file = fopen(config_file, "rt");
 	if ( file == NULL ) 
 		return ERROR;	/* return error designation. */
 	if ( header != NULL )
@@ -236,16 +338,17 @@ int	input_config(const char *filename, struct Config_Tag configs[], char *header
 
 			tok = trim(strtok(line, "=\n\r"));	 /* get first token */
 			if ( tok != NULL ) {
-				next = trim(strtok(NULL, "=\n\r")); /* get actual config information */
+				next = trim(strtok(NULL, "\n\r")); /* get actual config information */
 				for ( ptr = configs; ptr->buf; ++ptr )	 /* scan for token */ {
 					if ( !strcasecmp( tok , ptr->code ) )  /* got a match? */ {
 						if (next == NULL) {
-							if ( ptr->type == String_Tag ) {/* string may be empty */
+							if ( ptr->type == Path_Tag ||
+							     ptr->type == String_Tag ) {/* path or string may be empty */
 								*(char *)ptr->buf = 0;
 								++count;
 							}
 							else {
-								fprintf(stderr, ">>> Missing value in Config file %s on line %d !!!\n", filename, lineno);
+								panicbug(">>> Missing value in Config file %s on line %d !!!", config_file, lineno);
 							}
 							continue;
 						}
@@ -311,24 +414,32 @@ int	input_config(const char *filename, struct Config_Tag configs[], char *header
 							++count;
 							break;
 
-						case String_Tag:
+						case Path_Tag:
 							if (ptr->buf_size > 0) {
-								char *cptr = (char *)ptr->buf;
-								int bufsize = ptr->buf_size;
-								strncpy(cptr, next, bufsize);
-								cptr[bufsize-1] = '\0';	/* EOS */
+								char tmpbuf[MAX_PATH];
+								safe_strncpy(tmpbuf, next, sizeof(tmpbuf));
+								expand_path((char *)ptr->buf, tmpbuf, ptr->buf_size);
 								++count;
 							}
 							else {
-								fprintf(stderr, ">>> Wrong buf_size in Config_Tag struct: directive %s, buf_size %d !!!\n", ptr->code, ptr->buf_size);
+								panicbug(">>> Wrong buf_size in Config_Tag struct: directive %s, buf_size %d !!!", ptr->code, ptr->buf_size);
 							}
 							break;
 
+						case String_Tag:
+							if (ptr->buf_size > 0) {
+								safe_strncpy((char *)ptr->buf, next, ptr->buf_size);
+								++count;
+							}
+							else {
+								panicbug(">>> Wrong buf_size in Config_Tag struct: directive %s, buf_size %d !!!", ptr->code, ptr->buf_size);
+							}
+							break;
 						case Function_Tag:
 						case Error_Tag:
 						default:
 							printf("Error in Config file %s on line %d\n",
-								filename, lineno);
+								config_file, lineno);
 							break;
 						}
 					}
@@ -339,7 +450,22 @@ int	input_config(const char *filename, struct Config_Tag configs[], char *header
 	return count;
 }
 
-bool write_token(FILE *outfile, struct Config_Tag *ptr)
+
+int ConfigOptions::process_config(struct Config_Tag *conf, const char *title, bool verbose)
+{
+	int status = input_config(conf, title);
+	if (status >= 0) {
+		if (verbose)
+			infoprint("%s configuration: found %d valid directives.", title, status);
+	}
+	else {
+		panicbug("Error while reading/processing the '%s' config file.", config_file);
+	}
+	return status;
+}
+
+
+bool ConfigOptions::write_token(FILE *outfile, struct Config_Tag *ptr)
 {
 	int temp;
 	bool ret_flag = true;
@@ -354,7 +480,7 @@ bool write_token(FILE *outfile, struct Config_Tag *ptr)
 
 	case Byte_Tag:
 		temp = (int)*((char *)(ptr->buf));
-		fprintf(outfile, "%hd\n", temp);
+		fprintf(outfile, "%hd\n", (short)temp);
 		break;
 
 	case Word_Tag:
@@ -397,6 +523,14 @@ bool write_token(FILE *outfile, struct Config_Tag *ptr)
 		fprintf(outfile, "%c\n", *(char *)ptr->buf);
 		break;
 
+	case Path_Tag:
+		{
+			char tmpbuf[MAX_PATH];
+			compress_path(tmpbuf, (char *)ptr->buf, sizeof(tmpbuf));
+			fprintf(outfile, "%s\n", tmpbuf);
+		}
+		break;
+
 	case String_Tag:
 		fprintf(outfile, "%s\n", (char *)ptr->buf);
 		break;
@@ -417,12 +551,16 @@ bool write_token(FILE *outfile, struct Config_Tag *ptr)
 / return value:
 /	  int					  ; Number of records read & updated
 / parameters:
-/	  char *filename		  ; filename of INI file
 /	  struct Config_Tag configs[]; Configuration structure
 /	  char *header			  ; INI header name (i.e. "[TEST]")
 /-------------------------------------------------------------------<<*/
-int	update_config(const char *filename, struct Config_Tag configs[], char *header)
+int	ConfigOptions::update_config(struct Config_Tag configs[], const char *header)
 {
+#ifdef OS_darwin
+	static const char *tempfilename = "/tmp/aratemp.$$$";
+#else
+	static const char *tempfilename = "temp.$$$";
+#endif
 	struct Config_Tag *ptr;
 	int	count = 0, lineno = 0;
 	FILE * infile, *outfile;
@@ -432,12 +570,14 @@ int	update_config(const char *filename, struct Config_Tag configs[], char *heade
 	for ( ptr = configs; ptr->buf; ++ptr )
 		ptr->stat = 0;	/* jeste neulozeno do souboru */
 
-	infile = fopen(filename, "rt");
+	infile = fopen(config_file, "rt");
 	if ( infile == NULL ) {
 /* konfiguracni soubor jeste vubec neexistuje */
-		outfile = fopen("temp.$$$", "wt");
-		if ( outfile == NULL )
+		outfile = fopen(tempfilename, "wt");
+		if ( outfile == NULL ) {
+			panicbug("Error: unable to open %s file.", tempfilename);
 			return ERROR;		/* return error designation. */
+		}
 		if ( header != NULL ) {
 			fprintf(outfile, "%s\n", header);
 		}
@@ -447,15 +587,18 @@ int	update_config(const char *filename, struct Config_Tag configs[], char *heade
 		}
 
 		fclose(outfile);
-		result = fcopy(filename, "temp.$$$");
-		remove("temp.$$$");
+		result = fcopy(config_file, tempfilename);
+		remove(tempfilename);
 
-		if (result < 0)
+		if (result < 0) {
+			panicbug("Error %d in fcopy.", result);
 			return result;
+		}
 		return count;
 	}
-	outfile = fopen("temp.$$$", "wt");
+	outfile = fopen(tempfilename, "wt");
 	if ( outfile == NULL ) {
+		panicbug("Error: unable to open %s file.", tempfilename);
 		fclose(infile);
 		return ERROR;		   /* return error designation. */
 	}
@@ -529,11 +672,14 @@ int	update_config(const char *filename, struct Config_Tag configs[], char *heade
 	}
 	fclose(infile);
 	fclose(outfile);
-	result = fcopy(filename, "temp.$$$");
-	remove("temp.$$$");
+	result = fcopy(config_file, tempfilename);
+	remove(tempfilename);
 
-	if (result < 0)
+	if (result < 0) {
+		panicbug("Error %d in fcopy(%s,%s).", result, config_file, 
+				tempfilename);
 		return result;
+	}
 	return count;
 }
 
@@ -597,12 +743,13 @@ int	main(int argc, char *argv[])
 	printf("Options are now:\ntest1 = %s\ntest2 = %s\ntest3 = %d\n"
 		"test4 = %ld\ntest5 = \"%s\"\n\n", TFprint(test1),
 		TFprint(test2), test3, test4, test5);
-#endif
+#endif /* TEST_UPDATE */
 
 	return 0;
 }
 
-#endif
-
-} // extern "C"
+#endif /* TEST */
 
+/*
+vim:ts=4:sw=4:
+*/
