@@ -142,7 +142,7 @@ MD5::MD5Final(md5byte digest[16])
 
 	byteSwap(ctx->buf, 4);
 	memcpy(digest, ctx->buf, 16);
-	memset(ctx, 0, sizeof(ctx));	/* In case it's sensitive */
+	memset(ctx, 0, sizeof(MD5Context));	/* In case it's sensitive */
 }
 
 #ifndef ASM_MD5
