@@ -1,7 +1,7 @@
 /*
  * fpu_mpfr.cpp - emulate 68881/68040 fpu with mpfr
  *
- * Copyright (c) 2012 Andreas Schwab
+ * Copyright (c) 2012, 2013 Andreas Schwab
  *
  * ARAnyM is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -109,6 +109,15 @@ get_cur_prec ()
     }
 }
 
+#define DEFAULT_NAN_BITS 0xffffffffffffffffULL
+
+static void
+set_nan (fpu_register &reg, uae_u64 nan_bits)
+{
+  mpfr_set_nan (reg.f);
+  reg.nan_bits = nan_bits;
+}
+
 void
 fpu_init (bool integral_68040)
 {
@@ -173,7 +182,7 @@ fpu_reset ()
   fpu.instruction_address = 0;
 
   for (int i = 0; i < 8; i++)
-    mpfr_set_nan (fpu.registers[i].f);
+    set_nan (fpu.registers[i], DEFAULT_NAN_BITS);
 }
 
 fpu_register::operator long double ()
@@ -239,7 +248,7 @@ get_fp_addr (uae_u32 opcode, uae_u32 *addr, bool write)
 }
 
 static void
-set_from_single (mpfr_t value, uae_u32 data)
+set_from_single (fpu_register &value, uae_u32 data)
 {
   int s = data >> 31;
   int e = (data >> 23) & 0xff;
@@ -248,9 +257,13 @@ set_from_single (mpfr_t value, uae_u32 data)
   if (e == 0xff)
     {
       if (m != 0)
-	mpfr_set_nan (value);
+	{
+	  if (!(m & 0x400000))
+	    cur_exceptions |= FPSR_EXCEPTION_SNAN;
+	  set_nan (value, (uae_u64) (m | 0xc00000) << (32 + 8));
+	}
       else
-	mpfr_set_inf (value, 0);
+	mpfr_set_inf (value.f, 0);
     }
   else
     {
@@ -261,13 +274,13 @@ set_from_single (mpfr_t value, uae_u32 data)
 	e++;
       // Remove bias
       e -= SINGLE_BIAS;
-      mpfr_set_ui_2exp (value, m, e - (SINGLE_PREC - 1), MPFR_RNDN);
+      mpfr_set_ui_2exp (value.f, m, e - (SINGLE_PREC - 1), MPFR_RNDN);
     }
-  mpfr_setsign (value, value, s, MPFR_RNDN);
+  mpfr_setsign (value.f, value.f, s, MPFR_RNDN);
 }
 
 static void
-set_from_double (mpfr_t value, uae_u32 words[2])
+set_from_double (fpu_register &value, uae_u32 words[2])
 {
   int s = words[0] >> 31;
   int e = (words[0] >> 20) & 0x7ff;
@@ -275,10 +288,15 @@ set_from_double (mpfr_t value, uae_u32 words[2])
 
   if (e == 0x7ff)
     {
-      if (m != 0)
-	mpfr_set_nan (value);
+      if ((m | words[1]) != 0)
+	{
+	  if (!(m & 0x80000))
+	    cur_exceptions |= FPSR_EXCEPTION_SNAN;
+	  set_nan (value, (((uae_u64) (m | 0x180000) << (32 + 11))
+			   | ((uae_u64) words[1] << 11)));
+	}
       else
-	mpfr_set_inf (value, 0);
+	mpfr_set_inf (value.f, 0);
     }
   else
     {
@@ -289,14 +307,14 @@ set_from_double (mpfr_t value, uae_u32 words[2])
 	e++;
       // Remove bias
       e -= DOUBLE_BIAS;
-      mpfr_set_uj_2exp (value, ((uintmax_t) m << 32) | words[1],
+      mpfr_set_uj_2exp (value.f, ((uintmax_t) m << 32) | words[1],
 			e - (DOUBLE_PREC - 1), MPFR_RNDN);
     }
-  mpfr_setsign (value, value, s, MPFR_RNDN);
+  mpfr_setsign (value.f, value.f, s, MPFR_RNDN);
 }
 
 static void
-set_from_extended (mpfr_t value, uae_u32 words[3])
+set_from_extended (fpu_register &value, uae_u32 words[3], bool check_snan)
 {
   int s = words[0] >> 31;
   int e = (words[0] >> 16) & 0x7fff;
@@ -304,24 +322,32 @@ set_from_extended (mpfr_t value, uae_u32 words[3])
   if (e == 0x7fff)
     {
       if (((words[1] & 0x7fffffff) | words[2]) != 0)
-	mpfr_set_nan (value);
+	{
+	  if (check_snan)
+	    {
+	      if ((words[1] & 0x40000000) == 0)
+		cur_exceptions |= FPSR_EXCEPTION_SNAN;
+	      words[1] |= 0x40000000;
+	    }
+	  set_nan (value, ((uae_u64) words[1] << 32) | words[2]);
+	}
       else
-	mpfr_set_inf (value, 0);
+	mpfr_set_inf (value.f, 0);
     }
   else
     {
       // Remove bias
       e -= EXTENDED_BIAS;
-      mpfr_set_uj_2exp (value, ((uintmax_t) words[1] << 32) | words[2],
+      mpfr_set_uj_2exp (value.f, ((uintmax_t) words[1] << 32) | words[2],
 			e - (EXTENDED_PREC - 1), MPFR_RNDN);
     }
-  mpfr_setsign (value, value, s, MPFR_RNDN);
+  mpfr_setsign (value.f, value.f, s, MPFR_RNDN);
 }
 
 #define from_bcd(d) ((d) < 10 ? (d) : (d) - 10)
 
 static void
-set_from_packed (mpfr_t value, uae_u32 words[3])
+set_from_packed (fpu_register &value, uae_u32 words[3])
 {
   char str[32], *p = str;
   int sm = words[0] >> 31;
@@ -331,9 +357,13 @@ set_from_packed (mpfr_t value, uae_u32 words[3])
   if (((words[0] >> 16) & 0x7fff) == 0x7fff)
     {
       if ((words[1] | words[2]) != 0)
-	mpfr_set_nan (value);
+	{
+	  if ((words[1] & 0x40000000) == 0)
+	    cur_exceptions |= FPSR_EXCEPTION_SNAN;
+	  set_nan (value, ((uae_u64) (words[1] | 0x40000000) << 32) | words[2]);
+	}
       else
-	mpfr_set_inf (value, 0);
+	mpfr_set_inf (value.f, 0);
     }
   else
     {
@@ -354,13 +384,13 @@ set_from_packed (mpfr_t value, uae_u32 words[3])
       *p++ = from_bcd ((words[0] >> 20) & 15) + '0';
       *p++ = from_bcd ((words[0] >> 16) & 15) + '0';
       *p = 0;
-      mpfr_set_str (value, str, 10, MPFR_RNDN);
+      mpfr_set_str (value.f, str, 10, MPFR_RNDN);
     }
-  mpfr_setsign (value, value, sm, MPFR_RNDN);
+  mpfr_setsign (value.f, value.f, sm, MPFR_RNDN);
 }
 
 static bool
-get_fp_value (uae_u32 opcode, uae_u32 extra, mpfr_t value)
+get_fp_value (uae_u32 opcode, uae_u32 extra, fpu_register &value)
 {
   int mode, reg, size;
   uaecptr pc;
@@ -371,7 +401,14 @@ get_fp_value (uae_u32 opcode, uae_u32 extra, mpfr_t value)
 
   if ((extra & 0x4000) == 0)
     {
-      mpfr_set (value, fpu.registers[(extra >> 10) & 7].f, MPFR_RNDN);
+      mpfr_set (value.f, fpu.registers[(extra >> 10) & 7].f, MPFR_RNDN);
+      value.nan_bits = fpu.registers[(extra >> 10) & 7].nan_bits;
+      /* Check for SNaN.  */
+      if (mpfr_nan_p (value.f) && (value.nan_bits & (1ULL << 62)) == 0)
+	{
+	  value.nan_bits |= 1ULL << 62;
+	  cur_exceptions |= FPSR_EXCEPTION_SNAN;
+	}
       return true;
     }
   mode = (opcode >> 3) & 7;
@@ -383,13 +420,13 @@ get_fp_value (uae_u32 opcode, uae_u32 extra, mpfr_t value)
       switch (size)
 	{
 	case 6:
-	  mpfr_set_si (value, (uae_s8) m68k_dreg (regs, reg), MPFR_RNDN);
+	  mpfr_set_si (value.f, (uae_s8) m68k_dreg (regs, reg), MPFR_RNDN);
 	  break;
 	case 4:
-	  mpfr_set_si (value, (uae_s16) m68k_dreg (regs, reg), MPFR_RNDN);
+	  mpfr_set_si (value.f, (uae_s16) m68k_dreg (regs, reg), MPFR_RNDN);
 	  break;
 	case 0:
-	  mpfr_set_si (value, (uae_s32) m68k_dreg (regs, reg), MPFR_RNDN);
+	  mpfr_set_si (value.f, (uae_s32) m68k_dreg (regs, reg), MPFR_RNDN);
 	  break;
 	case 1:
 	  set_from_single (value, m68k_dreg (regs, reg));
@@ -444,7 +481,7 @@ get_fp_value (uae_u32 opcode, uae_u32 extra, mpfr_t value)
   switch (size)
     {
     case 0:
-      mpfr_set_si (value, (uae_s32) get_long (addr), MPFR_RNDN);
+      mpfr_set_si (value.f, (uae_s32) get_long (addr), MPFR_RNDN);
       break;
     case 1:
       set_from_single (value, get_long (addr));
@@ -453,7 +490,7 @@ get_fp_value (uae_u32 opcode, uae_u32 extra, mpfr_t value)
       words[0] = get_long (addr);
       words[1] = get_long (addr + 4);
       words[2] = get_long (addr + 8);
-      set_from_extended (value, words);
+      set_from_extended (value, words, true);
       break;
     case 3:
       words[0] = get_long (addr);
@@ -462,7 +499,7 @@ get_fp_value (uae_u32 opcode, uae_u32 extra, mpfr_t value)
       set_from_packed (value, words);
       break;
     case 4:
-      mpfr_set_si (value, (uae_s16) get_word (addr), MPFR_RNDN);
+      mpfr_set_si (value.f, (uae_s16) get_word (addr), MPFR_RNDN);
       break;
     case 5:
       words[0] = get_long (addr);
@@ -470,7 +507,7 @@ get_fp_value (uae_u32 opcode, uae_u32 extra, mpfr_t value)
       set_from_double (value, words);
       break;
     case 6:
-      mpfr_set_si (value, (uae_s8) get_byte (addr), MPFR_RNDN);
+      mpfr_set_si (value.f, (uae_s8) get_byte (addr), MPFR_RNDN);
       break;
     default:
       return false;
@@ -528,40 +565,12 @@ update_exceptions ()
 }
 
 static void
-set_fp_register (int reg, mpfr_t value, mpfr_prec_t prec, mpfr_rnd_t rnd, bool do_flags)
+set_fp_register (int reg, mpfr_t value, uae_u64 nan_bits,
+		 int t, mpfr_rnd_t rnd, bool do_flags)
 {
-  int t;
-
-  switch (prec)
-    {
-    case SINGLE_PREC:
-      {
-	MPFR_DECL_INIT (single, SINGLE_PREC);
-	// Round to single
-	set_format (SINGLE_PREC);
-	t = mpfr_set (single, value, rnd);
-	t = mpfr_check_range (single, t, rnd);
-	mpfr_subnormalize (single, t, rnd);
-	set_format (EXTENDED_PREC);
-	mpfr_set (fpu.registers[reg].f, single, rnd);
-      }
-      break;
-    case DOUBLE_PREC:
-      {
-	MPFR_DECL_INIT (dbl, DOUBLE_PREC);
-	// Round to double
-	set_format (DOUBLE_PREC);
-	t = mpfr_set (dbl, value, rnd);
-	t = mpfr_check_range (dbl, t, rnd);
-	mpfr_subnormalize (dbl, t, rnd);
-	set_format (EXTENDED_PREC);
-	mpfr_set (fpu.registers[reg].f, dbl, rnd);
-      }
-      break;
-    case EXTENDED_PREC:
-      mpfr_set (fpu.registers[reg].f, value, rnd);
-      break;
-    }
+  mpfr_subnormalize (value, t, rnd);
+  mpfr_set (fpu.registers[reg].f, value, rnd);
+  fpu.registers[reg].nan_bits = nan_bits;
   if (do_flags)
     {
       uae_u32 flags = 0;
@@ -579,7 +588,7 @@ set_fp_register (int reg, mpfr_t value, mpfr_prec_t prec, mpfr_rnd_t rnd, bool d
 }
 
 static uae_u32
-extract_to_single (mpfr_t value)
+extract_to_single (fpu_register &value)
 {
   uae_u32 word;
   int t;
@@ -588,7 +597,7 @@ extract_to_single (mpfr_t value)
 
   set_format (SINGLE_PREC);
   // Round to single
-  t = mpfr_set (single, value, rnd);
+  t = mpfr_set (single, value.f, rnd);
   t = mpfr_check_range (single, t, rnd);
   mpfr_subnormalize (single, t, rnd);
   set_format (EXTENDED_PREC);
@@ -596,7 +605,14 @@ extract_to_single (mpfr_t value)
   if (mpfr_inf_p (single))
     word = 0x7f800000;
   else if (mpfr_nan_p (single))
-    word = 0x7fffffff;
+    {
+      if ((value.nan_bits & (1ULL << 62)) == 0)
+	{
+	  value.nan_bits |= 1ULL << 62;
+	  cur_exceptions |= FPSR_EXCEPTION_SNAN;
+	}
+      word = 0x7f800000 | ((value.nan_bits >> (32 + 8)) & 0x7fffff);
+    }
   else if (mpfr_zero_p (single))
     word = 0;
   else
@@ -629,7 +645,7 @@ extract_to_single (mpfr_t value)
 }
 
 static void
-extract_to_double (mpfr_t value, uint32_t *words)
+extract_to_double (fpu_register &value, uint32_t *words)
 {
   int t;
   mpfr_rnd_t rnd = get_cur_rnd ();
@@ -637,7 +653,7 @@ extract_to_double (mpfr_t value, uint32_t *words)
 
   set_format (DOUBLE_PREC);
   // Round to double
-  t = mpfr_set (dbl, value, rnd);
+  t = mpfr_set (dbl, value.f, rnd);
   t = mpfr_check_range (dbl, t, rnd);
   mpfr_subnormalize (dbl, t, rnd);
   set_format (EXTENDED_PREC);
@@ -649,8 +665,13 @@ extract_to_double (mpfr_t value, uint32_t *words)
     }
   else if (mpfr_nan_p (dbl))
     {
-      words[0] = 0x7fffffff;
-      words[1] = 0xffffffff;
+      if ((value.nan_bits & (1ULL << 62)) == 0)
+	{
+	  value.nan_bits |= 1ULL << 62;
+	  cur_exceptions |= FPSR_EXCEPTION_SNAN;
+	}
+      words[0] = 0x7ff00000 | ((value.nan_bits >> (32 + 11)) & 0xfffff);
+      words[1] = value.nan_bits >> 11;
     }
   else if (mpfr_zero_p (dbl))
     {
@@ -689,21 +710,21 @@ extract_to_double (mpfr_t value, uint32_t *words)
 }
 
 static void
-extract_to_extended (mpfr_t value, uint32_t *words)
+extract_to_extended (fpu_register &value, uint32_t *words)
 {
-  if (mpfr_inf_p (value))
+  if (mpfr_inf_p (value.f))
     {
       words[0] = 0x7fff0000;
       words[1] = 0;
       words[2] = 0;
     }
-  else if (mpfr_nan_p (value))
+  else if (mpfr_nan_p (value.f))
     {
       words[0] = 0x7fff0000;
-      words[1] = 0xffffffff;
-      words[2] = 0xffffffff;
+      words[1] = value.nan_bits >> 32;
+      words[2] = value.nan_bits;
     }
-  else if (mpfr_zero_p (value))
+  else if (mpfr_zero_p (value.f))
     {
       words[0] = 0;
       words[1] = 0;
@@ -717,7 +738,7 @@ extract_to_extended (mpfr_t value, uint32_t *words)
       mpz_init (f);
       words[0] = words[1] = words[2] = 0;
       // Get exponent and mantissa
-      e = mpfr_get_z_2exp (f, value);
+      e = mpfr_get_z_2exp (f, value.f);
       // Move binary point
       e += EXTENDED_PREC - 1;
       // Add bias
@@ -735,26 +756,26 @@ extract_to_extended (mpfr_t value, uint32_t *words)
       words[0] = e << 16;
       mpz_clear (f);
     }
-  if (mpfr_signbit (value))
+  if (mpfr_signbit (value.f))
     words[0] |= 0x80000000;
 }
 
 static void
-extract_to_packed (mpfr_t value, int k, uae_u32 *words)
+extract_to_packed (fpu_register &value, int k, uae_u32 *words)
 {
-  if (mpfr_inf_p (value))
+  if (mpfr_inf_p (value.f))
     {
       words[0] = 0x7fff0000;
       words[1] = 0;
       words[2] = 0;
     }
-  else if (mpfr_nan_p (value))
+  else if (mpfr_nan_p (value.f))
     {
       words[0] = 0x7fff0000;
-      words[1] = 0xffffffff;
-      words[2] = 0xffffffff;
+      words[1] = value.nan_bits >> 32;
+      words[2] = value.nan_bits;
     }
-  else if (mpfr_zero_p (value))
+  else if (mpfr_zero_p (value.f))
     {
       words[0] = 0;
       words[1] = 0;
@@ -775,14 +796,14 @@ extract_to_packed (mpfr_t value, int k, uae_u32 *words)
 	{
 	  MPFR_DECL_INIT (temp, 16);
 
-	  mpfr_log10 (temp, value, rnd);
+	  mpfr_log10 (temp, value.f, rnd);
 	  k = mpfr_get_si (temp, MPFR_RNDZ) - k + 1;
 	}
       if (k <= 0)
 	k = 1;
       else if (k >= 18)
 	k = 17;
-      mpfr_get_str (str, &e, 10, k, value, rnd);
+      mpfr_get_str (str, &e, 10, k, value.f, rnd);
       e--;
       if (*p == '-')
 	p++;
@@ -810,7 +831,7 @@ extract_to_packed (mpfr_t value, int k, uae_u32 *words)
 	words[2] = (words[2] << 4) | (*p++ & 15);
 	  
     }  
-  if (mpfr_signbit (value))
+  if (mpfr_signbit (value.f))
     words[0] |= 0x80000000;
 }
 
@@ -852,7 +873,6 @@ fpuop_fmove_memory (uae_u32 opcode, uae_u32 extra)
   uaecptr pc;
   uae_u32 addr;
   uae_u32 words[3];
-  mpfr_t *value;
   static const int sz1[8] = {4, 4, 12, 12, 2, 8, 1, 0};
   static const int sz2[8] = {4, 4, 12, 12, 2, 8, 2, 0};
 
@@ -861,7 +881,7 @@ fpuop_fmove_memory (uae_u32 opcode, uae_u32 extra)
   mode = (opcode >> 3) & 7;
   reg = opcode & 7;
   size = (extra >> 10) & 7;
-  value = &fpu.registers[(extra >> 7) & 7].f;
+  fpu_register &value = fpu.registers[(extra >> 7) & 7];
 
   switch (mode)
     {
@@ -869,18 +889,18 @@ fpuop_fmove_memory (uae_u32 opcode, uae_u32 extra)
       switch (size)
 	{
 	case 0:
-	  m68k_dreg (regs, reg) = extract_to_integer (*value, -0x7fffffff-1, 0x7fffffff);
+	  m68k_dreg (regs, reg) = extract_to_integer (value.f, -0x7fffffff-1, 0x7fffffff);
 	  break;
 	case 1:
-	  m68k_dreg (regs, reg) = extract_to_single (*value);
+	  m68k_dreg (regs, reg) = extract_to_single (value);
 	  break;
 	case 4:
 	  m68k_dreg (regs, reg) &= ~0xffff;
-	  m68k_dreg (regs, reg) |= extract_to_integer (*value, -32768, 32767) & 0xffff;
+	  m68k_dreg (regs, reg) |= extract_to_integer (value.f, -32768, 32767) & 0xffff;
 	  break;
 	case 6:
 	  m68k_dreg (regs, reg) &= ~0xff;
-	  m68k_dreg (regs, reg) |= extract_to_integer (*value, -128, 127) & 0xff;
+	  m68k_dreg (regs, reg) |= extract_to_integer (value.f, -128, 127) & 0xff;
 	  break;
 	default:
 	  return false;
@@ -933,36 +953,36 @@ fpuop_fmove_memory (uae_u32 opcode, uae_u32 extra)
   switch (size)
     {
     case 0:
-      put_long (addr, extract_to_integer (*value, -0x7fffffff-1, 0x7fffffff));
+      put_long (addr, extract_to_integer (value.f, -0x7fffffff-1, 0x7fffffff));
       break;
     case 1:
-      put_long (addr, extract_to_single (*value));
+      put_long (addr, extract_to_single (value));
       break;
     case 2:
-      extract_to_extended (*value, words);
+      extract_to_extended (value, words);
       put_long (addr, words[0]);
       put_long (addr + 4, words[1]);
       put_long (addr + 8, words[2]);
       break;
     case 3:
-      extract_to_packed (*value, extra & 0x7f, words);
+      extract_to_packed (value, extra & 0x7f, words);
       put_long (addr, words[0]);
       put_long (addr + 4, words[1]);
       put_long (addr + 8, words[2]);
       break;
     case 4:
-      put_word (addr, extract_to_integer (*value, -32768, 32767));
+      put_word (addr, extract_to_integer (value.f, -32768, 32767));
       break;
     case 5:
-      extract_to_double (*value, words);
+      extract_to_double (value, words);
       put_long (addr, words[0]);
       put_long (addr + 4, words[1]);
       break;
     case 6:
-      put_byte (addr, extract_to_integer (*value, -128, 127));
+      put_byte (addr, extract_to_integer (value.f, -128, 127));
       break;
     case 7:
-      extract_to_packed (*value, m68k_dreg (regs, (extra >> 4) & 7) & 0x7f, words);
+      extract_to_packed (value, m68k_dreg (regs, (extra >> 4) & 7) & 0x7f, words);
       put_long (addr, words[0]);
       put_long (addr + 4, words[1]);
       put_long (addr + 8, words[2]);
@@ -1138,6 +1158,7 @@ fpuop_fmovem_register (uae_u32 opcode, uae_u32 extra)
   int list;
   int i;
 
+  set_format (EXTENDED_PREC);
   if (!get_fp_addr (opcode, &addr, extra & 0x2000))
     return false;
   if (extra & 0x800)
@@ -1159,7 +1180,7 @@ fpuop_fmovem_register (uae_u32 opcode, uae_u32 extra)
 	  for (i = 7; i >= 0; i--)
 	    if (list & (1 << i))
 	      {
-		extract_to_extended (fpu.registers[i].f, words);
+		extract_to_extended (fpu.registers[i], words);
 		addr -= 12;
 		put_long (addr, words[0]);
 		put_long (addr + 4, words[1]);
@@ -1173,7 +1194,7 @@ fpuop_fmovem_register (uae_u32 opcode, uae_u32 extra)
 	  for (i = 0; i < 8; i++)
 	    if (list & (0x80 >> i))
 	      {
-		extract_to_extended (fpu.registers[i].f, words);
+		extract_to_extended (fpu.registers[i], words);
 		put_long (addr, words[0]);
 		put_long (addr + 4, words[1]);
 		put_long (addr + 8, words[2]);
@@ -1200,7 +1221,7 @@ fpuop_fmovem_register (uae_u32 opcode, uae_u32 extra)
 	    words[1] = get_long (addr + 4);
 	    words[2] = get_long (addr + 8);
 	    addr += 12;
-	    set_from_extended (fpu.registers[i].f, words);
+	    set_from_extended (fpu.registers[i], words, false);
 	  }
       if ((opcode & 070) == 030)
 	m68k_areg (regs, opcode & 7) = addr;
@@ -1208,19 +1229,22 @@ fpuop_fmovem_register (uae_u32 opcode, uae_u32 extra)
   return true;
 }
 
-static void
-do_getexp (mpfr_t value)
+static int
+do_getexp (mpfr_t value, mpfr_rnd_t rnd)
 {
+  int t = 0;
+
   if (mpfr_inf_p (value))
     {
       mpfr_set_nan (value);
       cur_exceptions |= FPSR_EXCEPTION_OPERR;
     }
   else if (!mpfr_nan_p (value) && !mpfr_zero_p (value))
-    mpfr_set_si (value, mpfr_get_exp (value) - 1, MPFR_RNDN);
+    t = mpfr_set_si (value, mpfr_get_exp (value) - 1, rnd);
+  return t;
 }
 
-static void
+static int
 do_getman (mpfr_t value)
 {
   if (mpfr_inf_p (value))
@@ -1230,12 +1254,14 @@ do_getman (mpfr_t value)
     }
   else if (!mpfr_nan_p (value) && !mpfr_zero_p (value))
     mpfr_set_exp (value, 1);
+  return 0;
 }
 
-static void
+static int
 do_scale (mpfr_t value, mpfr_t reg, mpfr_rnd_t rnd)
 {
   long scale;
+  int t = 0;
 
   if (mpfr_nan_p (value))
     ;
@@ -1248,27 +1274,30 @@ do_scale (mpfr_t value, mpfr_t reg, mpfr_rnd_t rnd)
     {
       scale = mpfr_get_si (value, MPFR_RNDZ);
       mpfr_clear_inexflag ();
-      mpfr_mul_2si (value, reg, scale, rnd);
+      t = mpfr_mul_2si (value, reg, scale, rnd);
     }
   else
     mpfr_set_inf (value, -mpfr_signbit (value));
+  return t;
 }
 
-static void
+static int
 do_remainder (mpfr_t value, mpfr_t reg, mpfr_rnd_t rnd)
 {
   long quo;
+  int t = 0;
 
   if (mpfr_nan_p (value) || mpfr_nan_p (reg))
     ;
   else if (mpfr_zero_p (value) || mpfr_inf_p (reg))
     cur_exceptions |= FPSR_EXCEPTION_OPERR;
-  mpfr_remquo (value, &quo, reg, value, rnd);
+  t = mpfr_remquo (value, &quo, reg, value, rnd);
   if (quo < 0)
     quo = (-quo & 0x7f) | 0x80;
   else
     quo &= 0x7f;
   fpu.fpsr.quotient = quo << 16;
+  return t;
 }
 
 // Unfortunately, mpfr_fmod does not return the quotient bits, so we
@@ -1358,9 +1387,11 @@ mpfr_rem1 (mpfr_t rem, int *quo, mpfr_t x, mpfr_t y, mpfr_rnd_t rnd)
   return inex;
 }
 
-static void
+static int
 do_fmod (mpfr_t value, mpfr_t reg, mpfr_rnd_t rnd)
 {
+  int t = 0;
+
   if (mpfr_nan_p (value) || mpfr_nan_p (reg))
     mpfr_set_nan (value);
   else if (mpfr_zero_p (value) || mpfr_inf_p (reg))
@@ -1371,15 +1402,16 @@ do_fmod (mpfr_t value, mpfr_t reg, mpfr_rnd_t rnd)
   else if (mpfr_zero_p (reg) || mpfr_inf_p (value))
     {
       fpu.fpsr.quotient = 0;
-      mpfr_set (value, reg, rnd);
+      t = mpfr_set (value, reg, rnd);
     }
   else
     {
       int quo;
 
-      mpfr_rem1 (value, &quo, reg, value, rnd);
+      t = mpfr_rem1 (value, &quo, reg, value, rnd);
       fpu.fpsr.quotient = quo << 16;
     }
+  return t;
 }
 
 static void
@@ -1423,12 +1455,18 @@ do_ftst (mpfr_t value)
 static bool
 fpuop_general (uae_u32 opcode, uae_u32 extra)
 {
-  MPFR_DECL_INIT (value, EXTENDED_PREC);
   mpfr_prec_t prec = get_cur_prec ();
   mpfr_rnd_t rnd = get_cur_rnd ();
   int reg = (extra >> 7) & 7;
+  int t = 0;
+  fpu_register value;
+  bool ret;
+
+  mpfr_init2 (value.f, prec);
+  value.nan_bits = DEFAULT_NAN_BITS;
 
   mpfr_clear_flags ();
+  set_format (prec);
   cur_exceptions = 0;
   cur_instruction_address = m68k_getpc () - 4;
   if ((extra & 0xfc00) == 0x5c00)
@@ -1436,12 +1474,12 @@ fpuop_general (uae_u32 opcode, uae_u32 extra)
       // FMOVECR
       int rom_index = extra & 0x7f;
       if (rom_index == 0 || (rom_index >= 11 && rom_index <= 15))
-	mpfr_set (value, fpu_constant_rom[rom_index], rnd);
+	t = mpfr_set (value.f, fpu_constant_rom[rom_index], rnd);
       else if (rom_index >= 48 && rom_index <= 63)
-	mpfr_set (value, fpu_constant_rom[rom_index - 32], rnd);
+	t = mpfr_set (value.f, fpu_constant_rom[rom_index - 32], rnd);
       else
-	mpfr_set_zero (value, 0);
-      set_fp_register (reg, value, prec, rnd, true);
+	mpfr_set_zero (value.f, 0);
+      set_fp_register (reg, value.f, value.nan_bits, t, rnd, true);
     }
   else if (extra & 0x40)
     {
@@ -1455,103 +1493,120 @@ fpuop_general (uae_u32 opcode, uae_u32 extra)
 	  1, 0, 0, 0, 1, 0, 0, 0
 	};
 
-      if (!fpu.is_integral)
-	return false;
-      if (!valid[extra & 0x3b])
-	return false;
-      if (!get_fp_value (opcode, extra, value))
-	return false;
-
-      switch (extra & 0x3f)
-	{
-	case 0: // FSMOVE
-	case 4: // FDMOVE
-	  break;
-	case 1: // FSSQRT
-	case 5: // FDSQRT
-	  if (mpfr_sgn (value) < 0)
-	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_sqrt (value, value, rnd);
-	  break;
-	case 24: // FSABS
-	case 28: // FDABS
-	  mpfr_abs (value, value, rnd);
-	  break;
-	case 26: // FSNEG
-	case 30: // FDNEG
-	  mpfr_neg (value, value, rnd);
-	  break;
-	case 32: // FSDIV
-	case 36: // FDDIV
-	  if (mpfr_zero_p (value))
-	    {
-	      if (mpfr_regular_p (fpu.registers[reg].f))
-		cur_exceptions |= FPSR_EXCEPTION_DZ;
-	      else if (mpfr_zero_p (fpu.registers[reg].f))
-		cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	    }
-	  else if (mpfr_inf_p (value) && mpfr_inf_p (fpu.registers[reg].f))
-		cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_div (value, fpu.registers[reg].f, value, rnd);
-	  break;
-	case 34: // FSADD
-	case 38: // FDADD
-	  if (mpfr_inf_p (fpu.registers[reg].f) && mpfr_inf_p (value)
-	      && mpfr_signbit (fpu.registers[reg].f) != mpfr_signbit (value))
-	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_add (value, fpu.registers[reg].f, value, rnd);
-	  break;
-	case 35: // FSMUL
-	case 39: // FDMUL
-	  if ((mpfr_zero_p (value) && mpfr_inf_p (fpu.registers[reg].f))
-	      || (mpfr_inf_p (value) && mpfr_zero_p (fpu.registers[reg].f)))
-	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_mul (value, fpu.registers[reg].f, value, rnd);
-	  break;
-	case 40: // FSSUB
-	case 44: // FDSUB
-	  if (mpfr_inf_p (fpu.registers[reg].f) && mpfr_inf_p (value)
-	      && mpfr_signbit (fpu.registers[reg].f) == mpfr_signbit (value))
-	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_sub (value, fpu.registers[reg].f, value, rnd);
-	  break;
-	default:
-	  return false;
-	}
       if (extra & 4)
 	// FD...
 	prec = DOUBLE_PREC;
       else
 	// FS...
 	prec = SINGLE_PREC;
-      set_fp_register (reg, value, prec, rnd, true);
+      set_format (prec);
+      MPFR_DECL_INIT (value2, prec);
+
+      if (!fpu.is_integral)
+	{
+	  ret = false;
+	  goto out;
+	}
+      if (!valid[extra & 0x3b])
+	{
+	  ret = false;
+	  goto out;
+	}
+      if (!get_fp_value (opcode, extra, value))
+	{
+	  ret = false;
+	  goto out;
+	}
+
+      switch (extra & 0x3f)
+	{
+	case 0: // FSMOVE
+	case 4: // FDMOVE
+	  mpfr_set (value2, value.f, rnd);
+	  break;
+	case 1: // FSSQRT
+	case 5: // FDSQRT
+	  if (mpfr_sgn (value.f) < 0)
+	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
+	  t = mpfr_sqrt (value2, value.f, rnd);
+	  break;
+	case 24: // FSABS
+	case 28: // FDABS
+	  t = mpfr_abs (value2, value.f, rnd);
+	  break;
+	case 26: // FSNEG
+	case 30: // FDNEG
+	  t = mpfr_neg (value2, value.f, rnd);
+	  break;
+	case 32: // FSDIV
+	case 36: // FDDIV
+	  if (mpfr_zero_p (value.f))
+	    {
+	      if (mpfr_regular_p (fpu.registers[reg].f))
+		cur_exceptions |= FPSR_EXCEPTION_DZ;
+	      else if (mpfr_zero_p (fpu.registers[reg].f))
+		cur_exceptions |= FPSR_EXCEPTION_OPERR;
+	    }
+	  else if (mpfr_inf_p (value.f) && mpfr_inf_p (fpu.registers[reg].f))
+		cur_exceptions |= FPSR_EXCEPTION_OPERR;
+	  t = mpfr_div (value2, fpu.registers[reg].f, value.f, rnd);
+	  break;
+	case 34: // FSADD
+	case 38: // FDADD
+	  if (mpfr_inf_p (fpu.registers[reg].f) && mpfr_inf_p (value.f)
+	      && mpfr_signbit (fpu.registers[reg].f) != mpfr_signbit (value.f))
+	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
+	  t = mpfr_add (value2, fpu.registers[reg].f, value.f, rnd);
+	  break;
+	case 35: // FSMUL
+	case 39: // FDMUL
+	  if ((mpfr_zero_p (value.f) && mpfr_inf_p (fpu.registers[reg].f))
+	      || (mpfr_inf_p (value.f) && mpfr_zero_p (fpu.registers[reg].f)))
+	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
+	  t = mpfr_mul (value2, fpu.registers[reg].f, value.f, rnd);
+	  break;
+	case 40: // FSSUB
+	case 44: // FDSUB
+	  if (mpfr_inf_p (fpu.registers[reg].f) && mpfr_inf_p (value.f)
+	      && mpfr_signbit (fpu.registers[reg].f) == mpfr_signbit (value.f))
+	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
+	  t = mpfr_sub (value2, fpu.registers[reg].f, value.f, rnd);
+	  break;
+	}
+      set_fp_register (reg, value2, value.nan_bits, t, rnd, true);
     }
   else if ((extra & 0x30) == 0x30)
     {
       if ((extra & 15) > 10 || (extra & 15) == 9)
-	return false;
+	{
+	  ret = false;
+	  goto out;
+	}
       if (!get_fp_value (opcode, extra, value))
-	return false;
+	{
+	  ret = false;
+	  goto out;
+	}
 
       if ((extra & 15) < 8)
 	{
 	  // FSINCOS
 	  int reg2 = extra & 7;
-	  MPFR_DECL_INIT (value2, EXTENDED_PREC);
+	  MPFR_DECL_INIT (value2, prec);
 
-	  if (mpfr_inf_p (value))
+	  if (mpfr_inf_p (value.f))
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_sin_cos (value, value2, value, rnd);
+	  t = mpfr_sin_cos (value.f, value2, value.f, rnd);
 	  if (reg2 != reg)
-	    set_fp_register (reg2, value2, prec, rnd, false);
-	  set_fp_register (reg, value, prec, rnd, true);
+	    set_fp_register (reg2, value2, value.nan_bits, t >> 2, rnd, false);
+	  set_fp_register (reg, value.f, value.nan_bits, t & 3, rnd, true);
 	}
       else if ((extra & 15) == 8)
 	// FCMP
-	do_fcmp (value, fpu.registers[reg].f);
+	do_fcmp (value.f, fpu.registers[reg].f);
       else
 	// FTST
-	do_ftst (value);
+	do_ftst (value.f);
     }
   else
     {
@@ -1565,187 +1620,206 @@ fpuop_general (uae_u32 opcode, uae_u32 extra)
 	  1 
 	};
       if (!valid[extra & 0x3f])
-	return false;
+	{
+	  ret = false;
+	  goto out;
+	}
       if (!get_fp_value (opcode, extra, value))
-	return false;
+	{
+	  ret = false;
+	  goto out;
+	}
 
       switch (extra & 0x3f)
 	{
 	case 0: // FMOVE
 	  break;
 	case 1: // FINT
-	  mpfr_rint (value, value, rnd);
+	  t = mpfr_rint (value.f, value.f, rnd);
 	  break;
 	case 2: // FSINH
-	  mpfr_sinh (value, value, rnd);
+	  t = mpfr_sinh (value.f, value.f, rnd);
 	  break;
 	case 3: // FINTRZ
-	  mpfr_rint (value, value, MPFR_RNDZ);
+	  t = mpfr_rint (value.f, value.f, MPFR_RNDZ);
 	  break;
 	case 4: // FSQRT
-	  if (mpfr_sgn (value) < 0)
+	  if (mpfr_sgn (value.f) < 0)
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_sqrt (value, value, rnd);
+	  t = mpfr_sqrt (value.f, value.f, rnd);
 	  break;
 	case 6: // FLOGNP1
-	  if (!mpfr_nan_p (value))
+	  if (!mpfr_nan_p (value.f))
 	    {
-	      int cmp = mpfr_cmp_si (value, -1);
+	      int cmp = mpfr_cmp_si (value.f, -1);
 	      if (cmp == 0)
 		cur_exceptions |= FPSR_EXCEPTION_DZ;
 	      else if (cmp < 0)
 		cur_exceptions |= FPSR_EXCEPTION_OPERR;
 	    }
-	  mpfr_log1p (value, value, rnd);
+	  t = mpfr_log1p (value.f, value.f, rnd);
 	  break;
 	case 8: // FETOXM1
-	  mpfr_expm1 (value, value, rnd);
+	  t = mpfr_expm1 (value.f, value.f, rnd);
 	  break;
 	case 9: // FTANH
-	  mpfr_tanh (value, value, rnd);
+	  t = mpfr_tanh (value.f, value.f, rnd);
 	  break;
 	case 10: // FATAN
-	  mpfr_atan (value, value, rnd);
+	  t = mpfr_atan (value.f, value.f, rnd);
 	  break;
 	case 12: // FASIN
-	  if (mpfr_cmpabs (value, FPU_CONSTANT_ONE) > 0)
+	  if (mpfr_cmpabs (value.f, FPU_CONSTANT_ONE) > 0)
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_asin (value, value, rnd);
+	  t = mpfr_asin (value.f, value.f, rnd);
 	  break;
 	case 13: // FATANH
-	  if (mpfr_cmpabs (value, FPU_CONSTANT_ONE) > 0)
+	  if (mpfr_cmpabs (value.f, FPU_CONSTANT_ONE) > 0)
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_atanh (value, value, rnd);
+	  t = mpfr_atanh (value.f, value.f, rnd);
 	  break;
 	case 14: // FSIN
-	  if (mpfr_inf_p (value))
+	  if (mpfr_inf_p (value.f))
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_sin (value, value, rnd);
+	  t = mpfr_sin (value.f, value.f, rnd);
 	  break;
 	case 15: // FTAN
-	  if (mpfr_inf_p (value))
+	  if (mpfr_inf_p (value.f))
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_tan (value, value, rnd);
+	  t = mpfr_tan (value.f, value.f, rnd);
 	  break;
 	case 16: // FETOX
-	  mpfr_exp (value, value, rnd);
+	  t = mpfr_exp (value.f, value.f, rnd);
 	  break;
 	case 17: // FTWOTOX
-	  mpfr_ui_pow (value, 2, value, rnd);
+	  t = mpfr_ui_pow (value.f, 2, value.f, rnd);
 	  break;
 	case 18: // FTENTOX
-	  mpfr_ui_pow (value, 10, value, rnd);
+	  t = mpfr_ui_pow (value.f, 10, value.f, rnd);
 	  break;
 	case 20: // FLOGN
-	  if (mpfr_zero_p (value))
+	  if (mpfr_zero_p (value.f))
 	    cur_exceptions |= FPSR_EXCEPTION_DZ;
-	  else if (mpfr_sgn (value) < 0)
+	  else if (mpfr_sgn (value.f) < 0)
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_log (value, value, rnd);
+	  t = mpfr_log (value.f, value.f, rnd);
 	  break;
 	case 21: // FLOG10
-	  if (mpfr_zero_p (value))
+	  if (mpfr_zero_p (value.f))
 	    cur_exceptions |= FPSR_EXCEPTION_DZ;
-	  else if (mpfr_sgn (value) < 0)
+	  else if (mpfr_sgn (value.f) < 0)
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_log10 (value, value, rnd);
+	  t = mpfr_log10 (value.f, value.f, rnd);
 	  break;
 	case 22: // FLOG2
-	  if (mpfr_zero_p (value))
+	  if (mpfr_zero_p (value.f))
 	    cur_exceptions |= FPSR_EXCEPTION_DZ;
-	  else if (mpfr_sgn (value) < 0)
+	  else if (mpfr_sgn (value.f) < 0)
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_log2 (value, value, rnd);
+	  t = mpfr_log2 (value.f, value.f, rnd);
 	  break;
 	case 24: // FABS
-	  mpfr_abs (value, value, rnd);
+	  t = mpfr_abs (value.f, value.f, rnd);
 	  break;
 	case 25: // FCOSH
-	  mpfr_cosh (value, value, rnd);
+	  t = mpfr_cosh (value.f, value.f, rnd);
 	  break;
 	case 26: // FNEG
-	  mpfr_neg (value, value, rnd);
+	  t = mpfr_neg (value.f, value.f, rnd);
 	  break;
 	case 28: // FACOS
-	  if (mpfr_cmpabs (value, FPU_CONSTANT_ONE) > 0)
+	  if (mpfr_cmpabs (value.f, FPU_CONSTANT_ONE) > 0)
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_acos (value, value, rnd);
+	  t = mpfr_acos (value.f, value.f, rnd);
 	  break;
 	case 29: // FCOS
-	  if (mpfr_inf_p (value))
+	  if (mpfr_inf_p (value.f))
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_cos (value, value, rnd);
+	  t = mpfr_cos (value.f, value.f, rnd);
 	  break;
 	case 30: // FGETEXP
-	  do_getexp (value);
+	  t = do_getexp (value.f, rnd);
 	  break;
 	case 31: // FGETMAN
-	  do_getman (value);
+	  t = do_getman (value.f);
 	  break;
 	case 32: // FDIV
-	  if (mpfr_zero_p (value))
+	  if (mpfr_zero_p (value.f))
 	    {
 	      if (mpfr_regular_p (fpu.registers[reg].f))
 		cur_exceptions |= FPSR_EXCEPTION_DZ;
 	      else if (mpfr_zero_p (fpu.registers[reg].f))
 		cur_exceptions |= FPSR_EXCEPTION_OPERR;
 	    }
-	  else if (mpfr_inf_p (value) && mpfr_inf_p (fpu.registers[reg].f))
+	  else if (mpfr_inf_p (value.f) && mpfr_inf_p (fpu.registers[reg].f))
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_div (value, fpu.registers[reg].f, value, rnd);
+	  t = mpfr_div (value.f, fpu.registers[reg].f, value.f, rnd);
 	  break;
 	case 33: // FMOD
-	  do_fmod (value, fpu.registers[reg].f, rnd);
+	  t = do_fmod (value.f, fpu.registers[reg].f, rnd);
 	  break;
 	case 34: // FADD
-	  if (mpfr_inf_p (fpu.registers[reg].f) && mpfr_inf_p (value)
-	      && mpfr_signbit (fpu.registers[reg].f) != mpfr_signbit (value))
+	  if (mpfr_inf_p (fpu.registers[reg].f) && mpfr_inf_p (value.f)
+	      && mpfr_signbit (fpu.registers[reg].f) != mpfr_signbit (value.f))
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_add (value, fpu.registers[reg].f, value, rnd);
+	  t = mpfr_add (value.f, fpu.registers[reg].f, value.f, rnd);
 	  break;
 	case 35: // FMUL
-	  if ((mpfr_zero_p (value) && mpfr_inf_p (fpu.registers[reg].f))
-	      || (mpfr_inf_p (value) && mpfr_zero_p (fpu.registers[reg].f)))
+	  if ((mpfr_zero_p (value.f) && mpfr_inf_p (fpu.registers[reg].f))
+	      || (mpfr_inf_p (value.f) && mpfr_zero_p (fpu.registers[reg].f)))
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_mul (value, fpu.registers[reg].f, value, rnd);
+	  t = mpfr_mul (value.f, fpu.registers[reg].f, value.f, rnd);
 	  break;
 	case 36: // FSGLDIV
-	  if (mpfr_zero_p (value))
-	    {
-	      if (mpfr_regular_p (fpu.registers[reg].f))
-		cur_exceptions |= FPSR_EXCEPTION_DZ;
-	      else if (mpfr_zero_p (fpu.registers[reg].f))
-		cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	    }
-	  else if (mpfr_inf_p (value) && mpfr_inf_p (fpu.registers[reg].f))
-		cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_div (value, fpu.registers[reg].f, value, rnd);
-	  prec = SINGLE_PREC;
+	  {
+	    MPFR_DECL_INIT (value2, SINGLE_PREC);
+
+	    set_format (SINGLE_PREC);
+	    if (mpfr_zero_p (value.f))
+	      {
+		if (mpfr_regular_p (fpu.registers[reg].f))
+		  cur_exceptions |= FPSR_EXCEPTION_DZ;
+		else if (mpfr_zero_p (fpu.registers[reg].f))
+		  cur_exceptions |= FPSR_EXCEPTION_OPERR;
+	      }
+	    else if (mpfr_inf_p (value.f) && mpfr_inf_p (fpu.registers[reg].f))
+	      cur_exceptions |= FPSR_EXCEPTION_OPERR;
+	    t = mpfr_div (value2, fpu.registers[reg].f, value.f, rnd);
+	    mpfr_set (value.f, value2, rnd);
+	  }
 	  break;
 	case 37: // FREM
-	  do_remainder (value, fpu.registers[reg].f, rnd);
+	  t = do_remainder (value.f, fpu.registers[reg].f, rnd);
 	  break;
 	case 38: // FSCALE
-	  do_scale (value, fpu.registers[reg].f, rnd);
+	  t = do_scale (value.f, fpu.registers[reg].f, rnd);
 	  break;
 	case 39: // FSGLMUL
-	  if ((mpfr_zero_p (value) && mpfr_inf_p (fpu.registers[reg].f))
-	      || (mpfr_inf_p (value) && mpfr_zero_p (fpu.registers[reg].f)))
-	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_mul (value, fpu.registers[reg].f, value, rnd);
-	  prec = SINGLE_PREC;
+	  {
+	    MPFR_DECL_INIT (value2, SINGLE_PREC);
+
+	    set_format (SINGLE_PREC);
+	    if ((mpfr_zero_p (value.f) && mpfr_inf_p (fpu.registers[reg].f))
+		|| (mpfr_inf_p (value.f) && mpfr_zero_p (fpu.registers[reg].f)))
+	      cur_exceptions |= FPSR_EXCEPTION_OPERR;
+	    t = mpfr_mul (value2, fpu.registers[reg].f, value.f, rnd);
+	    mpfr_set (value.f, value2, rnd);
+	  }
 	  break;
 	case 40: // FSUB
-	  if (mpfr_inf_p (fpu.registers[reg].f) && mpfr_inf_p (value)
-	      && mpfr_signbit (fpu.registers[reg].f) == mpfr_signbit (value))
+	  if (mpfr_inf_p (fpu.registers[reg].f) && mpfr_inf_p (value.f)
+	      && mpfr_signbit (fpu.registers[reg].f) == mpfr_signbit (value.f))
 	    cur_exceptions |= FPSR_EXCEPTION_OPERR;
-	  mpfr_sub (value, fpu.registers[reg].f, value, rnd);
+	  t = mpfr_sub (value.f, fpu.registers[reg].f, value.f, rnd);
 	  break;
 	}
-      set_fp_register (reg, value, prec, rnd, true);
+      set_fp_register (reg, value.f, value.nan_bits, t, rnd, true);
     }
   update_exceptions ();
-  return true;
+  ret = true;
+ out:
+  mpfr_clear (value.f);
+  return ret;
 }
 
 void
@@ -1837,7 +1911,11 @@ void
 fpuop_bcc (uae_u32 opcode, uaecptr pc, uae_u32 disp)
 {
   if (check_fp_cond (opcode))
-    m68k_setpc (pc + (uae_s16) disp);
+    {
+      if (!(opcode & (1 << 6)))
+	disp = (uae_s16) disp;
+      m68k_setpc (pc + disp);
+    }
 }
 
 void
